/****************************************************************************
*
*	  Module Filename:  Random_Sequence.cpp
*	  Module Name:      Random_Sequence
*	  Description:      Maintains sequences of random numbers and can provide 
*                       random numbers based on the underlying sequences.
*                       Does this in a way that is unaffected by other sub-programs
*                       calling rand().
*                      
*	  Author:           Chris Gordon-Smith
*
*     Copyright (C) 2005, 2006, 2008 Chris Gordon-Smith
*     Email: See http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License Version 2, as published 
*     by the Free Software Foundation.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*	  Amendment History:-
*
*         Version    Date                 Update Description
*         =======    ===========          ==================
*
*
*
*
****************************************************************************/
//  Comment out to disallow tracing this file
//#define  ENABLE_TRACE_FILE 

#include    "globldef.h"
#include    "Library_Headers.h"

#include  <gsl/gsl_rng.h>    // GNU Scientific library

#include    "Random_Sequence.h"

//----------------------------------------------------------------------------------------------------------------

vector<Random_Sequence*>*  Random_Sequence::List_Of_Instance_Ptrs_Ptr;
int  Random_Sequence::Number_Of_Sequences  = 0;
//----------------------------------------------------------------------------------------------------------

bool    Random_Sequence::Check_Portability()

//  Check whether the underlying runtime libraries used by the Random_Sequence class
//  are compatible with those used on the SimSoup develoipment environment (Open Suse 10.1).
//  Specifically, is the sequence of pseudo-random numbers the same.

{    
    const gsl_rng_type* Rng_Type = gsl_rng_default;
    gsl_rng* Gen_Ptr = gsl_rng_alloc (Rng_Type);
    int Seed = 62987;
    gsl_rng_set (Gen_Ptr, Seed);

    for (int ILoop = 0; ILoop < 1000; ILoop++)
    BEGIN
        gsl_rng_uniform_int(Gen_Ptr, INT_MAX);
    END
    unsigned int Rand1= gsl_rng_uniform_int(Gen_Ptr, INT_MAX);
    unsigned int Rand2= gsl_rng_uniform_int(Gen_Ptr, INT_MAX);
    unsigned int Rand3= gsl_rng_uniform_int(Gen_Ptr, INT_MAX);

    if ((Rand1 not_eq 150583950) or (Rand2 not_eq 1458576491) or  (Rand3 not_eq 152299687))
    THEN
        return false;
    ELSE
        return true;
    ENDIF
}
//---------------------------------------------------------------------------------------------------------------

void  Random_Sequence::Clear_All()

//  Delete all the Random Sequences and the list of sequences

{
    TRACE("Random_Sequence::Clear_All");
    if (Number_Of_Sequences > 0)
    THEN
        vector<Random_Sequence*>::iterator  Iter;
        for (Iter = List_Of_Instance_Ptrs_Ptr->begin();
            Iter != List_Of_Instance_Ptrs_Ptr->end();
            ++Iter)
        BEGIN
            DELETE_OBJECT(Random_Sequence, *Iter);
        END
    ENDIF
    DELETE_OBJECT(vector<Random_Sequence*>, List_Of_Instance_Ptrs_Ptr);
    Number_Of_Sequences = 0;
}
//---------------------------------------------------------------------------------------------------------------

void  Random_Sequence::Make(int Num_Sequences)

//  Create the specified number of Random_Sequence objects

{
    TRACE("Random_Sequence::Make");
    Assert(Number_Of_Sequences == 0, "Random_Sequence::Make - Attempt to setup sequences a second time. Sequences must only be setup once");

    Number_Of_Sequences = Num_Sequences;

    CREATE_OBJECT(vector<Random_Sequence*>, List_Of_Instance_Ptrs_Ptr = new vector<Random_Sequence*>);

    for(int ILoop = 0;
        ILoop < Num_Sequences;
        ILoop++)
    BEGIN   
        Random_Sequence* Sequence_Ptr;
        CREATE_OBJECT(Random_Sequence, Sequence_Ptr = new Random_Sequence(1));   
        List_Of_Instance_Ptrs_Ptr->push_back(Sequence_Ptr);       
    END
}
//----------------------------------------------------------------------------------------------------------------

Random_Sequence::Random_Sequence(unsigned long int Seed)

//  Constructor

//  Setup a cache of random numbers. The simulation will use numbers from the cache cyclically as it runs
//  This is not ideal, but is done in this version of SimSoup to avoid problems
//  with non-repeatability that were experienced, possibly due to problems in Lazarus / Free Pascal initialisation
//  during the migration to GTK from Lazarus
//  We can return to runtime use of a random number generator when the migration is complete

{
    T = gsl_rng_default;
    r = gsl_rng_alloc (T);

    gsl_rng_set (r, Seed);

    Current_Index = 0;
    Cache_Size = 250000;

    int ILoop;
    for (ILoop = 0;
         ILoop < Cache_Size;
         ILoop++)
    BEGIN
        int  Rand_Val = gsl_rng_uniform_int(r, INT_MAX);
                                                // This function returns a random (Unsigned) integer from 0 to n-1 
                                                // inclusive by scaling down and/or discarding samples from 
                                                // the generator r. All integers in the range [0,n-1] are 
                                                // produced with equal probability. For generators with a 
                                                // non-zero minimum value an offset is applied so that zero is 
                                                // returned with the correct probability.

                                                // INT_MAX is used above because SimSoup uses integers, so 
                                                // we work here with numbers that are no bigger than INT_MAX

        Sequence.push_back(Rand_Val);
    END
}
//----------------------------------------------------------------------------------------------------------------

void Random_Sequence::Set_Seed(unsigned int s)

//  Set the random sequence seed. 

{
    gsl_rng_set (r, s);     // This function initializes (or `seeds') the random number generator. If the 
                            // generator is seeded with the same value of s on two different runs, the same 
                            // stream of random numbers will be generated by successive calls to the routines 
                            // below. If different values of s are supplied, then the generated streams of random 
                            // numbers should be completely different. If the seed s is zero then the standard 
                            // seed from the original implementation is used instead. For example, the original 
                            // Fortran source code for the ranlux generator used a seed of 314159265, and so 
                            // choosing s equal to zero reproduces this when using gsl_rng_ranlux.
}
//-----------------------------------------------------------------------------------------------------------------

int Random_Sequence::Uniform(int Maximum)

//  Return a non-negative random number between 0 and Maximum - 1 

{    
    int  Ret_Val = static_cast <int> (Sequence[Current_Index] * ((double(Maximum))/ INT_MAX));
 
    Assert(Ret_Val >= 0, "Random_Sequence::Uniform - Return Value is negative.");

    Assert((Ret_Val < Maximum) or Maximum == 0, "Random_Sequence::Uniform - Return Value is greater than Maximum - 1, and Maximum is non-zero");

    Current_Index++;
    if (Current_Index >= Cache_Size)
    THEN
        Current_Index = 0;
    ENDIF
   return Ret_Val;
}
//----------------------------------------------------------------------------------------------------------------
double Random_Sequence::Uniform(double Maximum)

//  Return a random floating point between 0 and Maximum (excluding Maximum itself)

{   
    double Ret_Val = Maximum * (double(Uniform(INT_MAX)) / INT_MAX);
    return Ret_Val;
}








